#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 =====================================================
                    TYPEDEFS
 =====================================================
 */

 typedef struct Array_element {
     struct Array_element *next;
     struct Array_element *prev;
     void *data;
 } array_element;

 typedef struct Array {
     int32_t length;
     array_element *head;
     array_element *tail;
     int32_t contains_struct;
 } array;

 typedef struct Map_element {
     struct Map_element *next;
     struct Map_element *prev;
     void *key;
     void *value;
 } map_element;

 typedef struct Map {
     int32_t length;
     map_element *head;
     map_element *tail;
     int32_t contains_struct;
     int32_t is_nodepair_map;
 } map;

 typedef struct Struct_element {
 	struct Struct_element *next;
 	struct Struct_element *prev;
 	char *field;
 	void *value;
 } struct_element;

 typedef struct Struct {
 	struct_element *head;
 	struct_element *tail;
  int32_t length;
 } sstruct;

 typedef struct Edge {
   sstruct *src;
   sstruct *dst;
   sstruct *value;
 } edge;

 typedef struct Graph {
   map *m;
   array *node_list;
 } graph;

 typedef struct NodePair {
   sstruct *src;
   sstruct *dst;
 } nodepair;

/*
  =====================================================
                  FUNCTION DECLARATIONS
  =====================================================
  */

array* arr_init();
array* arr_set_contains_struct(array *a);
array_element* arr_append(array *l, void *data);
void* arr_get(array *l, int index);
array_element* arr_set(array *l, void *data, int index);
int32_t arr_contains(array *l, void *data);
int32_t arr_length(array *l);

char *string_concat(char *s1, char *s2);
int string_equals(char *s1, char *s2);
int string_length(char *s);

map* map_init();
map* map_set_contains_struct(map *m);
map_element* map_get_el(map *m, void *key);
void* map_get(map *m, void *key);
void* map_set(map *m, void *key, void* value);
int32_t map_contains(map *m, void *key);
array* map_get_keys(map *m);
int32_t map_length(map *m);

sstruct * struct_init();
struct_element * struct_add(sstruct *s, char *field, void *value);
void* struct_get(sstruct *s, char *field);
struct_element* struct_set(sstruct *s, char *field, void *value);
int32_t struct_equals_helper(sstruct *s1, sstruct *s2);
int32_t struct_equals(sstruct *s1, sstruct *s2);

edge *edge_init(sstruct *src, sstruct *dst, sstruct *value);
sstruct *edge_get_src(edge *e);
sstruct *edge_get_dst(edge *e);
sstruct *edge_get_val(edge *e);
sstruct *edge_set_src(edge *e, sstruct *src);
sstruct *edge_set_dst(edge *e, sstruct *dst);
sstruct *edge_set_val(edge *e, sstruct *val);

int32_t nodepair_equals(nodepair *a, nodepair *b);
graph *graph_init();
edge *graph_add_edge(graph *g, edge *e);
array *graph_get_edges(graph *g);
array *graph_get_nodes(graph *g);
int32_t graph_contains_node(graph *g, sstruct *n);

double double_of_int(int x);
int int_of_double(double x);

/*
 =====================================================
                    ARRAY FUNCTIONS
 =====================================================
 */

 //init array
 array* arr_init() {
     array* l = (array *) malloc(sizeof(array));
     l->length = 0;
     l->head = NULL;
     l->tail = NULL;
     l->contains_struct = 0;
     return l;
 }

// call this if the array contains a struct
array* arr_set_contains_struct(array *a) {
  a->contains_struct = 1;
  return a;
}

 //append element: can be any type
 array_element* arr_append(array *l, void *data) {
     array_element *le = (array_element*) malloc(sizeof(array_element));
     le->data = data;
     le->next = NULL;
     le->prev = l->tail;
     if(l->tail) {
         l->tail->next = le;
     }
     else { //empty array
         l->head = le;
     }
     l->tail = le;
     l->length += 1;

     return le;
 }

 //access element
 void* arr_get(array *l, int index) {
     if(index >= l->length) {
         return NULL;
     }
     array_element *curr = l->head;
     int count = 0;
     while(count < index) {
         curr = curr->next;
         count++;
     }
     return curr->data;
 }

 //set element
 array_element* arr_set(array *l, void *data, int index) {
     if(index >= l->length) {
         return NULL;
     }
     array_element *curr = l->head;
     int count = 0;
     while(count < index) {
         curr = curr->next;
         count++;
     }
     free(curr->data);
     curr->data = data;
     return curr;
 }

 int32_t arr_contains(array *l, void *data) {
   array_element *curr = l->head;
   while(curr != NULL) {
     if (
       (
         (
           data == curr->data ||
           strcmp(data, curr->data) == 0
         ) && !l->contains_struct
       ) ||
       (l->contains_struct && struct_equals(*(sstruct **)data, *(sstruct **)curr->data))
     ) {
       return 1;
     }
     curr = curr->next;
   }
   return 0;

 }

 int32_t arr_length(array *l) {
     return l->length;
 }

/*
 =====================================================
                    STRING FUNCTIONS
 =====================================================
 */

char *string_concat(char *s1, char *s2) {
	char *new = (char *) malloc(strlen(s1) + strlen(s2) + 1);
	strcpy(new, s1);
	strcat(new, s2);
	return new;
}

int string_equals(char *s1, char *s2) {
	return !(strcmp(s1, s2) == 0);
}

int string_length(char *s) {
	return strlen(s);
}

/*
=====================================================
                   MAP FUNCTIONS
=====================================================
*/

/* Initialize map  */
map* map_init() {
  map* m = (map *) malloc(sizeof(map));
  m->length = 0;
  m->head = NULL;
  m->tail = NULL;
  m->contains_struct = 0;
  return m;
}

map* map_set_contains_struct(map *m) {
  m->contains_struct = 1;
  return m;
}

/* Get map element with key, used as a helper function */
map_element* map_get_el(map *m, void *key) {
  map_element *curr = m->head;
  int count = 0;
  while (count < m->length) {
    if (
      (
        (
          *(int *)curr->key == *(int *)key ||   /* if int key */
          *(char *)curr->key == *(char *)key || /* if char key */
          strcmp(curr->key, key) == 0           /* if str key */
        ) && !m->contains_struct && !m->is_nodepair_map
      ) ||
      (m->contains_struct && struct_equals(*(sstruct **)key, *(sstruct **)curr->key)) || /* if struct key */
      (m->is_nodepair_map && nodepair_equals((nodepair *)key, (nodepair *)curr->key))    /* if nodepair key */
    ) {
      break;
    }
    curr = curr->next;
    count++;
  }
  if (count == m->length) {
    // Not found
    return NULL;
  } else {
    // Found, return value
    return curr;
  }
}

/* Get value with int key */
void* map_get(map *m, void *key) {
  map_element *element = map_get_el(m, key);
  if (element == NULL) {
    return 0;
  } else {
    return element->value;
  }
}

/* Set value with int key */
void* map_set(map *m, void *key, void* value) {
  map_element *element = map_get_el(m, key);
  if (element == NULL) {
    // create new element and return
    map_element *new_element = malloc(sizeof(map_element));
    // printf("Malloc: at %p\n", new_element);
    // init new element
    new_element->key = key;
    new_element->value = value;
    new_element->prev = m->tail;
    new_element->next = NULL;
    // update pointers
    if (m->tail) {
      m->tail->next = new_element;
    } else {
      m->head = new_element;
    }
    m->tail = new_element;
    // increment
    m->length++;
  } else {
    // overwrite existing value
    element->value = value;
  }
	return value;
}

int32_t map_contains(map *m, void *key) {
  map_element *curr = m->head;
  while(curr != NULL) {
      if (
        (
          (
            *(int *)curr->key == *(int *)key ||   /* if int key */
            *(char *)curr->key == *(char *)key || /* if char key */
            strcmp(curr->key, key) == 0           /* if str key */
          ) && !m->contains_struct && !m->is_nodepair_map
        ) ||
        (m->contains_struct && struct_equals(*(sstruct **)key, *(sstruct **)curr->key)) || /* if struct key */
        (m->is_nodepair_map && nodepair_equals((nodepair *)key, (nodepair *)curr->key))    /* if nodepair key */
      ) {
        return 1;
      }
      curr = curr->next;
  }
  return 0;
}

array* map_get_keys(map *m) {
  array *a = arr_init();
  map_element *curr = m->head;
  while (curr != NULL) {
    arr_append(a, curr->key);
    curr = curr->next;
  }
  return a;
}

int32_t map_length(map *m) {
  return m->length;
}

/*
========================================================================
			Struct Functions
========================================================================
*/

sstruct * struct_init() {
	sstruct* s = (sstruct *) malloc(sizeof(sstruct));
	s->length = 0;
	s->head = NULL;
	s->tail = NULL;
	return s;
}

struct_element * struct_add(sstruct *s, char *field, void *value) {
	struct_element *s1 = (struct_element *) malloc(sizeof(struct_element));
	s1->field = field;
	s1->value = value;
	s1->prev = s->tail;
	if(s->tail){
		s->tail->next = s1;
	}
	else{
		s->head = s1;
	}
	s->tail = s1;
	s->length += 1;

	return s1;
}

void* struct_get(sstruct *s, char *field) {
	struct_element *cur = s->head;
	int count = 0;
	while(count < s->length){
		if (strcmp(cur->field, field) == 0){
			break;
		}
		else{
			cur = cur->next;
			count ++;
		}
	}
	return cur->value;
}

struct_element* struct_set(sstruct *s, char *field, void *value) {
	struct_element *cur = s->head;
	int count = 0;
	while(count < s->length){
		if (strcmp(cur->field, field) == 0){
			break;
		}
		else{
			cur = cur->next;
			count ++;
		}
	}
	free(cur->value);
	cur->value = value;
	return cur;
}

int32_t struct_equals_helper(sstruct *s1, sstruct *s2) {
  /* Since fields are potentially out of order, we'd need to validate each one. */
  struct_element *cur1 = s1->head;
  struct_element *cur2 = s2->head;
  while (cur1 != NULL) {
    while (cur2 != NULL && strcmp(cur1->field, cur2->field) != 0) {
      cur2 = cur2->next;
    }
    // if field is not found in s2
    if (cur2 == NULL) {
      return 0;
    }
    // now we know we're on the same field
    if (
      *(int *)cur1->value == *(int *)cur2->value || /* int */
      *(char *)cur1->value == *(char *)cur2->value || /* char */
      strcmp(cur1->value, cur2->value) == 0 /* string */
    ) {
      cur1 = cur1->next;
      continue;
    }
    // now we know fields don't equal
    return 0;
  }
  return 1;
}

int32_t struct_equals(sstruct *s1, sstruct *s2) {
  /* Sanity check on length */
  if (s1->length != s2->length) {
    return 0;
  }
  return struct_equals_helper(s1, s2) || struct_equals_helper(s2, s1);
}

/*
=====================================================
                   EDGE FUNCTIONS
=====================================================
*/

edge *edge_init(sstruct *src, sstruct *dst, sstruct *value) {
  edge *e = (edge *)malloc(sizeof(edge));
  e->src = src;
  e->dst = dst;
  e->value = value;
  return e;
}

sstruct *edge_get_src(edge *e) {
  return e->src;
}

sstruct *edge_get_dst(edge *e) {
  return e->dst;
}

sstruct *edge_get_val(edge *e) {
  return e->value;
}

sstruct *edge_set_src(edge *e, sstruct *src) {
  e->src = src;
  return src;
}

sstruct *edge_set_dst(edge *e, sstruct *dst) {
  e->dst = dst;
  return dst;
}

sstruct *edge_set_val(edge *e, sstruct *value) {
  e->value = value;
  return value;
}

/*
=====================================================
              GRAPH/NODEPAIR FUNCTIONS
=====================================================
*/

/*
  Implementation note: we use a map to store the graph, where the key type is
  nodepair (an ordered pair of nodes), and value is sstruct (the edge value),
  i.e. map<nodepair, sstruct>
*/

int32_t nodepair_equals(nodepair *a, nodepair *b) {
  return struct_equals(a->src, b->src) && struct_equals(a->dst, b->dst);
}

graph *graph_init() {
  graph *g = malloc(sizeof(graph));
  g->m = map_init();
  g->m->is_nodepair_map = 1;
  g->node_list = arr_init();
  g->node_list->contains_struct = 1;
  return g;
}

/* Node: this function overwrites an edge if it already exists */
edge *graph_add_edge(graph *g, edge *e) {
  nodepair *np = malloc(sizeof(nodepair));
  np->src = e->src;
  np->dst = e->dst;
  map_set(g->m, np, e->value);
  /* add src to node list */
  if (!arr_contains(g->node_list, &np->src)) {
    sstruct **ptr = malloc(sizeof(sstruct *));
    *ptr = np->src;
    arr_append(g->node_list, ptr);
  }
  if (!arr_contains(g->node_list, &np->dst)) {
    sstruct **ptr = malloc(sizeof(sstruct *));
    *ptr = np->dst;
    arr_append(g->node_list, ptr);
  }
  return e;
}

array *graph_get_edges(graph *g) {
  array *a = arr_init();
  map_element *curr = g->m->head;
  while (curr != NULL) {
    edge *e = edge_init(((nodepair *)curr->key)->src, ((nodepair *)curr->key)->dst, curr->value);
    edge **e_ptr = malloc(sizeof(edge *));
    *e_ptr = e;
    arr_append(a, e_ptr);
    curr = curr->next;
  }
  return a;
}

array *graph_get_nodes(graph *g) {
  return g->node_list;
}

int32_t graph_contains_node(graph *g, sstruct *n) {
  return arr_contains(g->node_list, n);
}

/*
==================================================
		Int to/from Double
==================================================
*/

double double_of_int(int x) {
	return ((double) x);
}

int int_of_double(double x) {
	return (int) x;
}
